/*

XFILTER - device driver which captures IDL output

xfilter.c

Copyright (C) 1999 Craig B. Markwardt.  All Rights Reserved.

This C code implements a graphics device "filter" for IDL.  The intent
is to capture graphics commands which can be reproduced at a later
time.  For example, to capture the graphics that appear on the X
windows display for printing later.

Currently this works only under X windows and Unix (Linux, Solaris,
SunOS, and Alpha are tested).  Graphics commands are captured and can
be recovered using the XFILTER IDL commands.

The method is simple.  A "new" graphics device is registered.  IDL's
notion of a graphics device is a jump table which implements various
drawing primitives.  Each primitive is implemented below, but
trivially calls the true X device primitive.  Thus the devices are
chained together.  Both graphics and windowing primitives are chained.

The non-portable aspects of this program can be found in xfiltdlm.c

IDL is a trademark of RSI.

This file is part of XFILTER.

XFILTER is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

XFILTER is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with XFILTER; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/*
  HOW IT WORKS

  Calls to IDL graphics routines are intercepted, and redirected to
  wrapper routines called "stubs" below.  The stubs are instantiated
  by the xfilter_insert_stub macro.

  Calls to IDL device routines are also intercepted, but only the
  ERASE function is redirected.  It is used to clear the cache.

  When a call to a graphics routine is made, the stub does the
  following:

  * saves the arguments with xfilter_prep_cmd;
  * calls the real IDL graphics command;
  * stores the arguments on a window-specific stack with xfilter_store_cmd.

  xfilter_prep_cmd() does the meat of the work.  It copies each of the
  arguments and each of the keywords.  There is a special case for the
  _EXTRA keyword.  There is a fair amount of machinery involved in
  copying the data, mostly because structures are involved.

  External IDL commands are registered which can query the state of,
  and retrieve, the command data stored for a particular window.  See
  PROGRAMMING for a discussion of that interface.

  Initialization is the worst part of the whole thing.  There are
  architecture-dependent symbols that must be found, and an
  architecture dependent way to determine structure tag names (for IDL
  < 5.2).  This is handled and documented in xfiltdlm.c.

*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "export.h"

extern char * xfilter_tag_name(IDL_StructDefPtr sdef, int tag_num);
extern void xfilter_init_symbols();

IDL_DEVICE_DEF    *px_dev;        /* IDL X windows device driver */
IDL_DEVICE_DEF    xfilter_dev;    /* New device driver block */
IDL_DEVICE_CORE   *xdev, *xfdev;
IDL_DEVICE_WINDOW *xwin, *xfwin;

/* Finite state machine states.  Some/most are obsolete now. */
#define		XFILTER_VERSION		"0.8"
#define		XFILTER_STARTUP		0
#define		XFILTER_WAIT_CMD	1
#define		XFILTER_WAIT_DEV	2
#define		XFILTER_GOOD_CMD	3
int		xfilter_state = 0;

/*********************************************************************/
/*                XFILTER Window Handling                            */

#define		XFILTER_MAX_WINDOWS	8
int		xfilter_win[XFILTER_MAX_WINDOWS];
int		xfilter_win_size[XFILTER_MAX_WINDOWS][2];
IDL_VARIABLE	*xfilter_kcache[XFILTER_MAX_WINDOWS];
IDL_VARIABLE	*xfilter_vcache[XFILTER_MAX_WINDOWS];
IDL_VARIABLE	*xfilter_ccache[XFILTER_MAX_WINDOWS];
int		xfilter_cache_max[XFILTER_MAX_WINDOWS];
int		xfilter_cache_num[XFILTER_MAX_WINDOWS];
int		xfilter_cache_stat[XFILTER_MAX_WINDOWS];
int		xfilter_cache_bytes[XFILTER_MAX_WINDOWS];
IDL_VARIABLE	xfilter_null;


/* Empty a window's in-memory cache */
int xfilter_win_uncache(int win)
{
  int i, j, wmax = XFILTER_MAX_WINDOWS;

  for(i=0; i<wmax; i++) {
    if (xfilter_win[i] == win && xfilter_cache_stat[i]) {
      for (j=0; j<xfilter_cache_num[i]; j++) {
	IDL_StoreScalarZero(&xfilter_kcache[i][j], IDL_TYP_LONG);
	IDL_StoreScalarZero(&xfilter_vcache[i][j], IDL_TYP_LONG);
	if (xfilter_ccache[i][j].type == IDL_TYP_STRING) {
	  IDL_StrDelete(&xfilter_ccache[i][j].value.str, 1);
	  xfilter_ccache[i][j].type  = IDL_TYP_UNDEF;
	  xfilter_ccache[i][j].flags = 0;
	}
	IDL_StoreScalarZero(&xfilter_ccache[i][j], IDL_TYP_LONG);
      }
      xfilter_cache_num[i]   = 0;
      xfilter_cache_bytes[i] = 0;
      break;
    }
  }
  return 0;
}

/* Initialize a window for recording */
int xfilter_win_init(int win)
{
  int i, wmax = XFILTER_MAX_WINDOWS;

  for(i=0; i<wmax; i++) {
    if (xfilter_win[i] == win) {
      xfilter_win_uncache(win);
      return 0;
    }
  }

  for(i=0; i<wmax; i++) {
    if (xfilter_win[i] < 0) {
      xfilter_win[i]         = win;
      xfilter_cache_num[i]   = 0;
      xfilter_cache_stat[i]  = 1;  /* Enable caching */
      xfilter_cache_bytes[i] = 0;
      xfilter_win_size[i][0] = 0;
      xfilter_win_size[i][1] = 0;
      return 1;
    }
  }
  return 0;
}

/* Remove a window from recording */
int xfilter_win_del(int win)
{
  int i, wmax = XFILTER_MAX_WINDOWS;

  xfilter_win_uncache(win);
  for(i=0; i<wmax; i++) {
    if (xfilter_win[i] == win) {
      xfilter_win[i]         = -1;
      xfilter_cache_num[i]   = 0;
      xfilter_cache_stat[i]  = 0;
      xfilter_cache_bytes[i] = 0;
      xfilter_win_size[i][0] = 0;
      xfilter_win_size[i][1] = 0;
      break;
    }
  }
  return 0;
}

/* Entry point for XFILTER_WINDOW */
void xfilter_window_proc(int argc, IDL_VPTR *argv)
{
  int win = IDL_LongScalar(argv[0]);
  if (win >= 0) xfilter_win_init(win);
  /* Error checking can be done by doing a N_COMMAND control afterward */
}

/* Entry point for XFILTER_WDELETE */
void xfilter_wdelete_proc(int argc, IDL_VPTR *argv)
{
  int win = IDL_LongScalar(argv[0]);
  if (win >= 0) xfilter_win_del(win);
}

#define OUT_KW (IDL_KW_ZERO|IDL_KW_OUT)

/* Entry point for XFILTER_HELP */
void xfilter_help_proc(int argc, IDL_VPTR *argv, char *argk)
{
  int i, first = 1;
  static IDL_VPTR version;
  static IDL_KW_PAR control_kw[] = {
    {"VERSION", IDL_TYP_UNDEF, 1, OUT_KW,      0, IDL_CHARA(version)},
    {0,         0,             0, 0,           0, 0                 }, 
  };
  IDL_VPTR pargv[1];

  /* Parse keywords */
  argc = IDL_KWGetParams(argc, argv, argk, control_kw, pargv, 1);

  /* Pull out version number if desired. */
  if (version) {
    IDL_VPTR cptr;
    cptr = IDL_StrToSTRING(XFILTER_VERSION);
    IDL_VarCopy(cptr, version);
    return;
  }    

  /* Print summary of window information */
  for (i=0; i<XFILTER_MAX_WINDOWS; i++) 
    if (xfilter_win[i] >= 0) {
      char buffer[200];
      if (first)
	IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO|IDL_MSG_ATTR_NOPREFIX,
		    "            XFILTERED WINDOW SUMMARY\n"
		    "  Window  # Cmds  X Size  Y Size     Bytes\n");
      sprintf(buffer, "  %4d     %4d    %4d    %4d  %8d\n", 
	     xfilter_win[i], xfilter_cache_num[i], 
	     xfilter_win_size[i][0], xfilter_win_size[i][1], 
	     xfilter_cache_bytes[i]);
      IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO|IDL_MSG_ATTR_NOPREFIX, buffer);
      first = 0;
    }
  if (first)
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO|IDL_MSG_ATTR_NOPREFIX, 
		"\n   NO WINDOWS XFILTERED\n\n");
  return;
}

/* Entry point for XFILTER_CONTROL */
/* CONTROL is the main procedure.  Clients can retrieve information
   about data in the graphics window, and update its state. */
void xfilter_control_proc(int argc, IDL_VPTR *argv, char *argk)
{
  static IDL_LONG start, stop, erase, read;
  static IDL_VPTR n_commands, xsize, ysize;
  static IDL_KW_PAR control_kw[] = {
    {"ERASE", IDL_TYP_LONG,  1, IDL_KW_ZERO, 0, IDL_CHARA(erase)},
    {"N_COMMANDS", 
              IDL_TYP_UNDEF, 1, OUT_KW,      0, IDL_CHARA(n_commands)},
    {"READ",  IDL_TYP_LONG,  1, IDL_KW_ZERO, 0, IDL_CHARA(read)},
    {"START", IDL_TYP_LONG,  1, IDL_KW_ZERO, 0, IDL_CHARA(start)},
    {"STOP",  IDL_TYP_LONG,  1, IDL_KW_ZERO, 0, IDL_CHARA(stop)},
    {"XSIZE", IDL_TYP_UNDEF, 1, OUT_KW,      0, IDL_CHARA(xsize)},
    {"YSIZE", IDL_TYP_UNDEF, 1, OUT_KW,      0, IDL_CHARA(ysize)},
    {0,       0,             0, 0,           0, 0},
  };
  IDL_VPTR pargv[5];
  int win, pos, i;
  IDL_ALLTYPES at;

  /* Parse keywords */
  argc = IDL_KWGetParams(argc, argv, argk, control_kw, pargv, 1);

  if (!read && argc > 1) {
    IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_LONGJMP,
		"Non-READ operations must have only one argument");
  }    
  
  /* Store number of commands, part 1 */
  if (n_commands) {
    at.l = -1;
    IDL_StoreScalar(n_commands, IDL_TYP_LONG, &at);
  }

  /* Establish the window number */
  IDL_ENSURE_SIMPLE(pargv[0]);
  IDL_EXCLUDE_STRING(pargv[0]);
  win  = IDL_LongScalar(pargv[0]);
  
  for (i=0; i<XFILTER_MAX_WINDOWS; i++) if (xfilter_win[i] == win) break;
  if (i == XFILTER_MAX_WINDOWS) i = -1;
      
  /* Handle flags */
  if (start && i >= 0) xfilter_cache_stat[i] = 1;
  if (stop  && i >= 0) xfilter_cache_stat[i] = 0;
  if (erase && i >= 0) xfilter_win_uncache(win);

  /* Store number of commands */
  if (n_commands) {
    if (i >= 0) at.l = xfilter_cache_num[i];
    else        at.l = -1;
    IDL_StoreScalar(n_commands, IDL_TYP_LONG, &at);
  }

  /* Various retrieval keywords */

  if (i >= 0) {
    /* Read window X size */
    if (xsize) {
      at.l = xfilter_win_size[i][0];
      IDL_StoreScalar(xsize, IDL_TYP_LONG, &at);
    }
    /* Read window Y size */
    if (ysize) {
      at.l = xfilter_win_size[i][1];
      IDL_StoreScalar(ysize, IDL_TYP_LONG, &at);
    }
  
    /* Perform read operation */
    if (read) {
      if (argc != 5) {
	IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_LONGJMP,
		    "Must give five arguments for READ operation");
      }
      
      IDL_ENSURE_SIMPLE(pargv[1]);
      IDL_EXCLUDE_STRING(pargv[1]);
      pos = IDL_LongScalar(pargv[1]);
      
      if (pos >=0 && pos < xfilter_cache_num[i]) {
	IDL_VarCopy(&xfilter_ccache[i][pos], pargv[2]);
	IDL_VarCopy(&xfilter_vcache[i][pos], pargv[3]);
	IDL_VarCopy(&xfilter_kcache[i][pos], pargv[4]);
      }
    }
  }

  return;
}

/* Enlarge the window cache */
void xfilter_win_boost(int win)
{
  int i, j, j0, wmax = XFILTER_MAX_WINDOWS;
  IDL_VARIABLE *vv;

  for(i=0; i<wmax; i++) {
    if (xfilter_win[i] == win) {
      j = j0 = xfilter_cache_max[i];
      if (j == 0)       j  = 8;
      else if (j < 128) j *= 2; 
      else              j += 128;
      vv = (IDL_VARIABLE *) IDL_MemAlloc(j*sizeof(IDL_VARIABLE), 
					 "XFILTER", IDL_MSG_EXIT);
      bzero((char *) vv, j*sizeof(IDL_VARIABLE));
      if (j0 > 0) {
	bcopy((char *) xfilter_vcache[i], (char *)vv, j0*sizeof(IDL_VARIABLE));
	IDL_MemFree(xfilter_vcache[i], "XFILTER", IDL_MSG_EXIT);
      }
      xfilter_vcache[i] = vv;
      vv = (IDL_VARIABLE *) IDL_MemAlloc(j*sizeof(IDL_VARIABLE), 
					 "XFILTER", IDL_MSG_EXIT);
      bzero((char *) vv, j*sizeof(IDL_VARIABLE));
      if (j0 > 0) {
	bcopy((char *) xfilter_kcache[i], (char *)vv, j0*sizeof(IDL_VARIABLE));
	IDL_MemFree(xfilter_kcache[i], "XFILTER", IDL_MSG_EXIT);
      }
      xfilter_kcache[i] = vv;
      vv = (IDL_VARIABLE *) IDL_MemAlloc(j*sizeof(IDL_VARIABLE), 
					 "XFILTER", IDL_MSG_EXIT);
      bzero((char *) vv, j*sizeof(IDL_VARIABLE));
      if (j0 > 0) {
	bcopy((char *) xfilter_ccache[i], (char *)vv, j0*sizeof(IDL_VARIABLE));
	IDL_MemFree(xfilter_ccache[i], "XFILTER", IDL_MSG_EXIT);
      }
      xfilter_ccache[i] = vv;
      xfilter_cache_max[i] = j;
      break;
    }
  }
}

/* Push another command onto a window's stack */
int xfilter_win_push(int win, IDL_VPTR cv, IDL_VPTR vv, IDL_VPTR kv)
{
  int i, j, wmax = XFILTER_MAX_WINDOWS;

  for(i=0; i<wmax; i++) {
    if (xfilter_win[i] == win && xfilter_cache_stat[i]) {
      /* Save size of window if the first time around */
      if (xfilter_cache_num[i] == 0) {
	xfilter_win_size[i][0] = px_dev->t_size[0];
	xfilter_win_size[i][1] = px_dev->t_size[1];
      }
      if (xfilter_cache_num[i] == xfilter_cache_max[i])
	xfilter_win_boost(win);
      if (vv->flags & IDL_V_ARR) 
	xfilter_cache_bytes[i] += vv->value.arr->arr_len;
      if (kv->flags & IDL_V_ARR) 
	xfilter_cache_bytes[i] += kv->value.arr->arr_len;
      j = xfilter_cache_num[i];
      xfilter_ccache[i][j] = *cv;
      xfilter_kcache[i][j] = *kv;
      xfilter_vcache[i][j] = *vv;
      xfilter_cache_num[i] ++;
      return 1;   /* SUCCESS */
    }
  }
  return 0;  /* FAILURE */
}

  
/*********************************************************************/
/*                IDL Built-in Procedure Chaining                    */


/* Hash table for built-in procedures */
extern long int **proc_table;
/* Find string in hash table */
extern IDL_SYSFUN_DEF *(*proc_search)(unsigned long int *, char *);

/* Chain a new procedure */
IDL_FUN_RET xfilter_chain(char *fname, IDL_FUN_RET newfun)
{
  char capfname[256], *p;
  IDL_SYSFUN_DEF *fdef;
  IDL_FUN_RET oldfun = 0;

  if (fname == 0 || fname[0] == 0 || newfun == 0) return 0;
  /* Capitalize name */
  strcpy(capfname, fname);
  for(p = capfname; *p; p++) *p = toupper((int) *p);
  /* Find name */
  fdef = (*proc_search)(*proc_table, capfname);
  if (fdef == 0) return 0;
  /* Chain function */
  oldfun = fdef->funct_addr;
  if (oldfun == 0) return 0;
  fdef->funct_addr = newfun;
  return oldfun;
}

/* Copy a variable's data to dptr */
void xfilter_copy_data(IDL_VPTR src, char *dptr)
{
  int slen = 0;
  char *sptr;

  if (src->flags & (IDL_V_STRUCT|IDL_V_ARR)) {
    sptr = (char *) src->value.arr->data;
    slen = src->value.arr->arr_len;
  } else if (src->type == IDL_TYP_STRING) {
    IDL_STRING *dst = (IDL_STRING *) dptr;
    *dst = src->value.str;
    IDL_StrDup(dst, 1);
    return;
  } else {
    sptr = (char *) &src->value;
    switch(src->type) {
    case IDL_TYP_BYTE     : slen =  1;  break;
    case IDL_TYP_INT      : slen =  2;  break;
    case IDL_TYP_LONG     : slen =  4;  break;
    case IDL_TYP_FLOAT    : slen =  4;  break;
    case IDL_TYP_DOUBLE   : slen =  8;  break;
    case IDL_TYP_COMPLEX  : slen =  8;  break;
    case IDL_TYP_DCOMPLEX : slen = 16;  break;
    case IDL_TYP_UINT     : slen =  2;  break;
    case IDL_TYP_ULONG    : slen =  4;  break;
    case IDL_TYP_LONG64   : slen =  8;  break;
    case IDL_TYP_ULONG64  : slen =  8;  break;
    }
  }

  if (slen > 0) bcopy((char *) sptr, (char *) dptr, slen);
}

/* Add a new tag to a structure */
int xfilter_new_tag(IDL_STRUCT_TAG_DEF *td, IDL_MEMINT xdims[][9], int ik, 
		     char *name, IDL_VPTR v, int ntags)
{
  int j;

  td[ik].name   = name;
  if (name[0] == '0' && name[1] == 0 && (v->flags & IDL_V_STRUCT)) {  
    /* Deal with _EXTRA keyword */
    for (j=0; j<ntags; j++) {
      IDL_VPTR vi;
      char *tname;

      /* Recursive call adds each element to the struct definition */
      IDL_StructTagInfoByIndex(v->value.s.sdef, j, 0, &vi);
      
      /* WARNING: this function was not available before IDL 5.2 */
      tname = xfilter_tag_name(v->value.s.sdef, j);
      if (tname == 0) {
	IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_EXIT,
		    "ERROR: Could not extract tag name.");
	exit(1);
      }
      ik = xfilter_new_tag(td, xdims, ik, tname, vi, 1);
    }
    return ik;
  } else if (v->flags & (IDL_V_STRUCT|IDL_V_ARR)) {
    xdims[ik][0] = v->value.arr->n_dim;
    for (j=0; j<xdims[ik][0]; j++) xdims[ik][j+1] = v->value.arr->dim[j];
    td[ik].dims = xdims[ik];
  } else
    td[ik].dims = 0;
  if (v->type == IDL_TYP_STRUCT)
    td[ik].type = v->value.s.sdef;
  else
    td[ik].type = (void *) ((unsigned long int) v->type) ;
  td[ik].flags  = 0;

  return (ik+1);
}

/* Create a new structure */
IDL_VPTR xfilter_make_struct(IDL_STRUCT_TAG_DEF *td, 
		    int argc, IDL_VPTR *argv, int *mask)
{
  char *dptr0;
  IDL_MEMINT one[] = {1};
  IDL_StructDefPtr ksd;
  IDL_VPTR kptr;
  int i, ik;

  /* Null string in the last element */
  ik = 0;
  for (i=0; i<argc; i++) if (mask[i] > 0) ik += mask[i];
  td[ik].name = 0;

  ksd = IDL_MakeStruct(0, td);
  if (ksd == 0) return 0;

  dptr0 = (char *) IDL_MakeTempStruct(ksd, 1, one, &kptr, 1);
  if (dptr0 == 0 || kptr == 0) return 0;

  ik = 0;
  for (i=0; i<argc; i++) {
    if (mask[i] > 0) {
      char *dptr;

      int doff = IDL_StructTagInfoByIndex(kptr->value.s.sdef, ik, 
					  IDL_MSG_RET, 0);
      if (doff < 0) {
	IDL_Deltmp(kptr);
	return 0;
      }
      dptr = dptr0 + doff;
      xfilter_copy_data(argv[i], dptr);
      ik += mask[i];
    }
  }
  return kptr;
}

/* Store copies of a command's data */
IDL_VARIABLE gc, gv, gk;
void xfilter_prep_cmd(int win, char *proname, 
		      int argc, IDL_VPTR *argv, char *argk)
{
  int i, ic = 0, ik = 0;
  IDL_STRUCT_TAG_DEF td[IDL_MAXPARAMS];
  IDL_MEMINT xdims[IDL_MAXPARAMS][9];
  int kmask[IDL_MAXPARAMS];
  char tagnames[IDL_MAXPARAMS][5];
  IDL_VPTR vptr, kptr, cptr;
  char *kp = argk;
  static int first = 1;
  
  if (first) {
    bzero((char *) &gc, sizeof(IDL_VARIABLE));
    bzero((char *) &gv, sizeof(IDL_VARIABLE));
    bzero((char *) &gk, sizeof(IDL_VARIABLE));
    first = 0;
  }
  if (gc.type != IDL_TYP_UNDEF) IDL_StoreScalarZero(&gc, IDL_TYP_LONG);
  if (gv.type != IDL_TYP_UNDEF) IDL_StoreScalarZero(&gv, IDL_TYP_LONG);
  if (gk.type != IDL_TYP_UNDEF) IDL_StoreScalarZero(&gk, IDL_TYP_LONG);

  for (i = 0; i<XFILTER_MAX_WINDOWS; i++)
    if (win == xfilter_win[i] && xfilter_cache_stat[i] == 1) goto goodwin;
  return;

 goodwin:
  if (proname)
    cptr = IDL_StrToSTRING(proname);
  else 
    cptr = &xfilter_null;
  
  for (i=0; i<argc; i++) 
    kmask[i] = 0;

  if (kp) {
    for (i=0; i<argc; i++) {
      if (*kp) {
	if (argv[i]->type != IDL_TYP_UNDEF) {
	  kmask[i] = 1;
	  if (kp[0] == '0' && kp[1] == 0)
	    kmask[i] = IDL_StructNumTags(argv[i]->value.s.sdef);
	  ik = xfilter_new_tag(td, xdims, ik, kp, argv[i], kmask[i]);
	}
	while (*kp) kp++;
      }
      kp ++;
    }
    if (ik) 
      kptr = xfilter_make_struct(td, argc, argv, kmask);
  }
  if (kp == 0 || ik == 0)
    kptr = &xfilter_null;
  
  /* Do regular args */
  for (i=0; i<argc; i++) {
    kmask[i] = 1 - (kmask[i]>0);
    if (kmask[i] > 0 && argv[i]->type == IDL_TYP_UNDEF) kmask[i] = 0;
    if (kmask[i] > 0) {
      if ((argv[i]->flags & IDL_V_ARR) != 0  
	  && argv[i]->value.arr->n_elts == 1)
	sprintf(tagnames[ic], "A%d", ic);
      else
	sprintf(tagnames[ic], "X%d", ic);
      ic = xfilter_new_tag(td, xdims, ic, tagnames[ic], argv[i], kmask[i]);
    }
  }
  if (ic) {
    vptr = xfilter_make_struct(td, argc, argv, kmask);
  } else {
    vptr = &xfilter_null;
  }
  IDL_VarCopy(cptr, &gc); /*  IDL_DELTMP(cptr); */
  IDL_VarCopy(kptr, &gk); /*  IDL_DELTMP(kptr); */
  IDL_VarCopy(vptr, &gv); /*  IDL_DELTMP(vptr); */
}

/* Store the command's data into the window's cache */
void xfilter_store_cmd(int win)
{
  if (gc.type && gc.type != IDL_TYP_LONG) {
    if (xfilter_win_push(win, &gc, &gv, &gk)) {
      /* Success: we zero out the variable */
      bzero((char *) &gc, sizeof(IDL_VARIABLE));
      bzero((char *) &gv, sizeof(IDL_VARIABLE));
      bzero((char *) &gk, sizeof(IDL_VARIABLE));
    } else {
      /* Failure: we release the storage of the variable */
      IDL_StoreScalarZero(&gc, IDL_TYP_LONG);
      IDL_StoreScalarZero(&gv, IDL_TYP_LONG);
      IDL_StoreScalarZero(&gk, IDL_TYP_LONG);
    }
  }
}

/*********************************************************************/
/*                Replacement IDL graphics routines                  */
#define xfilter_curwin (px_dev->window)
#define xfilter_curdev (IDL_PlotCom.dev)
#define xfilter_device_stub(fname)                                           \
IDL_VPTR (*xfilter_old_##fname)(int argc, IDL_VPTR *argv, char *argk);       \
IDL_VPTR xfilter_##fname(int argc, IDL_VPTR *argv, char *argk)               \
{                                                                            \
  IDL_VPTR retval;                                                           \
  xfilter_state = XFILTER_WAIT_DEV;                                          \
  if (xfilter_curdev == px_dev)                                              \
    xfilter_prep_cmd(xfilter_curwin, #fname, argc, argv, argk);              \
  retval = (*xfilter_old_##fname)(argc, argv, argk);                         \
  /* NOTE:  the above call can crash and skip the call below. */             \
  if (xfilter_curdev == px_dev)                                              \
    xfilter_store_cmd(xfilter_curwin);                                       \
  return retval;                                                             \
} extern int xfilter_dummy

#define xfilter_insert_stub(fname) \
  xfilter_old_##fname = xfilter_chain(#fname, xfilter_##fname)

xfilter_device_stub(axis);
xfilter_device_stub(contour);
/* xfilter_device_stub(erase); - don't record erase events! */
xfilter_device_stub(map_continents);
xfilter_device_stub(map_grid);
xfilter_device_stub(map_set);
xfilter_device_stub(oplot);
xfilter_device_stub(plot);
xfilter_device_stub(plot_io);
xfilter_device_stub(plot_oi);
xfilter_device_stub(plot_oo);
xfilter_device_stub(plots);
xfilter_device_stub(polyfill);
xfilter_device_stub(shade_surf);
xfilter_device_stub(shade_volume);
xfilter_device_stub(surface);
xfilter_device_stub(tv);
xfilter_device_stub(tvscl);
xfilter_device_stub(tvlct);
xfilter_device_stub(xyouts);

/*********************************************************************/
/*                     Graphics primitives                           */

void xfilter_check_state()
{
  int i, wmax = XFILTER_MAX_WINDOWS, win = xfilter_curwin;

  if (xfilter_state == XFILTER_WAIT_DEV) {
    for (i=0; i<wmax; i++)
      if (xfilter_win[i] == win && xfilter_cache_stat[i]) {
	xfilter_state = XFILTER_GOOD_CMD;
	break;
      }
  }
}

/* Draw line segment */
void xfilter_dev_draw(IDL_GR_PT *p0, IDL_GR_PT *p1, IDL_ATTR_STRUCT *a) 
{
  /*  xfilter_check_state(); */
  xdev->draw(p0, p1, a);   /* Call original X driver */
}

/* Draw hardware text */
int xfilter_dev_text(IDL_GR_PT *p0, IDL_ATTR_STRUCT *a, 
		 IDL_TEXT_STRUCT *ta, char *t)
{
  /* xfilter_check_state(); */
  return xdev->text(p0, a, ta, t);   /* Call original X driver */
}

/* Erase graphics viewport */
void xfilter_dev_erase(IDL_ATTR_STRUCT *a)
{
  int i, wmax = XFILTER_MAX_WINDOWS, win = xfilter_curwin;

  for (i=0; i<wmax; i++)
    if (xfilter_win[i] == win && xfilter_cache_stat[i])
      xfilter_win_uncache(xfilter_curwin);
  xdev->erase(a);   /* Call original X driver */
}

/* Capture cursor information */
void xfilter_dev_cursor(int funct, IDL_MOUSE_STRUCT *m)
{
  xdev->cursor(funct, m);   /* Call original X driver */
}

/* Filled polygons.  Yuck */
void xfilter_dev_polyfill(int *x, int *y, int n, IDL_POLYFILL_ATTR *poly)
{
  /* xfilter_check_state(); */
  xdev->polyfill(x, y, n, poly);   /* Call original X driver */
}

/* Called when returning to interactive level; graphics state should be
   well defined after this call.
 */
void xfilter_dev_inter_exit(void)
{
  xdev->inter_exit();   /* Call original X driver */
}

/* Flush any pending graphics output to the display */
void xfilter_dev_flush(void)
{
  xdev->flush();   /* Call original X driver */
}

/* Load new color map?  Doesn't seem to be used */
void xfilter_dev_load_color(IDL_LONG start, IDL_LONG n)
{
  xdev->load_color(start, n);   /* Call original X driver */
}

/* Read or write a pixmap to the display */
void xfilter_dev_rw_pixels(UCHAR *data, int x0, int y0, int nx,
		       int ny, int dir, IDL_TV_STRUCT *secondary)
{
  /* xfilter_check_state(); */
  /* Call original X driver */
  xdev->rw_pixels(data, x0, y0, nx, ny, dir, secondary);
}

/* Implement DEVICE command */
void xfilter_dev_dev_specific(int argc, IDL_VPTR *argv, char *argk)
{
  xdev->dev_specific(argc, argv, argk);   /* Call original X driver */
}

/* Implement HELP, /DEVICE */
void xfilter_dev_dev_help(int argc, IDL_VPTR *argv)
{
  xdev->dev_help(argc, argv);   /* Call original X driver */
}

/* Called once, when device driver is installed */
void xfilter_dev_load_rtn(void)
{
  /* Don't call the original driver, since it expects to be called once (?) */
  /* xdev->load_rtn(); */
}

/*********************************************************************/
/*                     Window   primitives                           */

/* Create Window */
void xfilter_window_create(int argc, IDL_VPTR *argv, char *argk)
{
  xwin->window_create(argc, argv, argk);
}

/* Delete Window */
void xfilter_window_delete(int argc, IDL_VPTR *argv)
{
  /* XXX must release cache here XXX */
  xwin->window_delete(argc, argv);
}

/* Show (activate) window */
void xfilter_window_show(int argc, IDL_VPTR *argv, char *argk)
{
  xwin->window_show(argc, argv, argk);
}

/* Set new active window */
void xfilter_window_set(int argc, IDL_VPTR *argv)
{
  xwin->window_set(argc, argv);
}  

/* Window menuing (?) */  
IDL_VPTR xfilter_window_menu(int argc, IDL_VPTR *argv, char *argk)
{
  return xwin->window_menu(argc, argv, argk);
}  

/*********************************************************************/
/*                          Initialization                           */

IDL_SYSFUN_DEF xfilter_proc_addr[] = {
  { (IDL_FUN_RET) xfilter_control_proc, "XFILTER_CONTROL", 1, 5, 
                                         IDL_SYSFUN_DEF_F_KEYWORDS},
  { (IDL_FUN_RET) xfilter_help_proc,    "XFILTER_HELP",    0, 0, 
                                         IDL_SYSFUN_DEF_F_KEYWORDS},
  { (IDL_FUN_RET) xfilter_window_proc,  "XFILTER_WINDOW",  1, 1, 0},
  { (IDL_FUN_RET) xfilter_wdelete_proc, "XFILTER_WDELETE", 1, 1, 0},
};

extern void xfilter_init_symbols();

long int xfilter_init(int argc, void *argv[])
{
  static int init = 0;
  int i;
  IDL_PLOT_COM *pc;

  if (init) return 0;

  /* Extract X device */
  pc = IDL_PlotComAddr();
  if (pc == 0) goto DEV_DEF_ERROR;

  /* Be sure this is really the X device */
  px_dev = pc->dev;
  if (px_dev == 0) goto DEV_DEF_ERROR;
  if (strcmp(IDL_STRING_STR(&px_dev->name), "X") != 0) {
  DEV_DEF_ERROR:
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, 
		"ERROR: Could not start XFILTER (X device was not present).");
    return 0;
  }

  xfilter_init_symbols();
  init = 1;
  xfilter_insert_stub(axis);
  xfilter_insert_stub(contour);
  /* xfilter_insert_stub(erase); */
  xfilter_insert_stub(map_continents);
  xfilter_insert_stub(map_grid);
  xfilter_insert_stub(map_set);
  xfilter_insert_stub(oplot);
  xfilter_insert_stub(plot);
  xfilter_insert_stub(plot_io);
  xfilter_insert_stub(plot_oi);
  xfilter_insert_stub(plot_oo);
  xfilter_insert_stub(plots);
  xfilter_insert_stub(polyfill);
  xfilter_insert_stub(shade_surf);
  xfilter_insert_stub(shade_volume);
  xfilter_insert_stub(surface);
  xfilter_insert_stub(tv);
  xfilter_insert_stub(tvscl);
  xfilter_insert_stub(tvlct);
  xfilter_insert_stub(xyouts);

  /* Copy device driver */
  xfilter_dev = *px_dev;

  /* Convenience pointers, makes things easier to type above */
  xdev  = &xfilter_dev.core;
  xfdev = &px_dev->core;
  xwin  = &xfilter_dev.winsys;
  xfwin = &px_dev->winsys;

  /* Set up new jump table - graphics primitives */
  xfdev->draw         = &xfilter_dev_draw;
  xfdev->text         = &xfilter_dev_text;
  xfdev->erase        = &xfilter_dev_erase;
  xfdev->cursor       = &xfilter_dev_cursor;
  xfdev->polyfill     = &xfilter_dev_polyfill;
  xfdev->inter_exit   = &xfilter_dev_inter_exit;
  xfdev->flush        = &xfilter_dev_flush;
  xfdev->rw_pixels    = &xfilter_dev_rw_pixels;
  xfdev->dev_specific = &xfilter_dev_dev_specific;
  xfdev->dev_help     = &xfilter_dev_dev_help;
  xfdev->load_rtn     = &xfilter_dev_load_rtn;

  /* Set up new jump table - window primitives */
  xfwin->window_create = &xfilter_window_create;
  xfwin->window_delete = &xfilter_window_delete;
  xfwin->window_show   = &xfilter_window_show;
  xfwin->window_set    = &xfilter_window_set;
  xfwin->window_menu   = &xfilter_window_menu;

  /* Add device */
  IDL_AddDevice(px_dev, IDL_MSG_EXIT);

  /* Initialize xfilter state */
  xfilter_state = XFILTER_WAIT_CMD;
  IDL_StoreScalarZero(&xfilter_null, IDL_TYP_LONG);

  /* Initialize window management */
  for (i=0; i<XFILTER_MAX_WINDOWS; i++) {
    xfilter_win[i] = -1;
    xfilter_cache_max[i] = 0;
    xfilter_cache_num[i] = 0;
  }
  
  /* Add system routines */
  IDL_AddSystemRoutine(xfilter_proc_addr, 0, 4);  /* Procedures */

  return 0;
}

long int *xfilter_check(int argc, void *argv[])
{
  int x;

  x = 0;
  
  return 0;
}
